"""
These classes are a collection of the needed tools to read external data.
The External type objects created by these classes are initialized before
the Stateful objects by functions.Model.initialize.
"""

import re
import warnings

from openpyxl import load_workbook
from openpyxl.utils.exceptions import InvalidFileException

import numpy as np
import xarray as xr
import pandas as pd

from . import utils
from .data import Data
from .lookups import Lookups


_SPREADSHEET_EXTS = {'.xls', '.xlsx', '.xlsm', '.xlsb', '.odf', '.ods', '.odt'}


class Excels():
    """
    Class to save the read Excel files and thus avoid double reading
    """
    _Excels, _Excels_opyxl = {}, {}

    @classmethod
    def read(cls, file_name, tab):
        """
        Read the Excel file or return the previously read one
        """
        if file_name.joinpath(tab) in cls._Excels:
            return cls._Excels[file_name.joinpath(tab)]
        else:
            # get the function to read the data based on its extension
            read_kwargs = {}
            ext = file_name.suffix.lower()
            if ext in _SPREADSHEET_EXTS:
                read_func = pd.read_excel
                read_kwargs['sheet_name'] = tab
            elif ext == '.csv':
                read_func = pd.read_csv
                if tab and not tab[0].isalnum():
                    read_kwargs['sep'] = tab
            else:
                read_func = pd.read_table
                if tab and not tab[0].isalnum():
                    read_kwargs['sep'] = tab
            # read the data
            excel = np.array([
                pd.to_numeric(ex, errors='coerce')
                for ex in
                read_func(file_name, header=None, **read_kwargs).values
                ])
            # save data for future retrievals
            cls._Excels[file_name.joinpath(tab)] = excel
            return excel

    @classmethod
    def read_opyxl(cls, file_name):
        """
        Read the Excel file using OpenPyXL or return the previously read one
        """
        if file_name in cls._Excels_opyxl:
            return cls._Excels_opyxl[file_name]
        else:
            excel = load_workbook(file_name, read_only=True, data_only=True)
            cls._Excels_opyxl[file_name] = excel
            return excel

    @classmethod
    def clean(cls):
        """
        Clean the dictionary of read files
        """
        for file in cls._Excels_opyxl.values():
            # close files open directly with openpyxls
            file.close()
            # files open with pandas are automatically closed
        cls._Excels, cls._Excels_opyxl = {}, {}


class External(object):
    """
    Main class of external objects

    Attributes
    ----------
    py_name: str
        The Python name of the object
    missing: str ("warning", "error", "ignore", "keep")
        What to do with missing values. If "warning" (default)
        shows a warning message and interpolates the values.
        If "raise" raises an error. If "ignore" interpolates
        the values without showing anything. If "keep" it will keep
        the missing values, this option may cause the integration to
        fail, but it may be used to check the quality of the data.
    file: str
        File name from which the data is read.
    tab: str
        Tab name from which the data is read. If file type is not a
        spreadsheet this will be used as a separator.

    """
    missing = "warning"

    def __init__(self, py_name):
        self.py_name = py_name
        self.file = None
        self.tab = None

    def __str__(self):
        return self.py_name

    def _get_data_from_file(self, rows, cols):
        """
        Function to read data from excel file using rows and columns

        Parameters
        ----------
        rows: list of len 2
            first row and last row+1 to be read, starting from 0
        cols:  list of len 2
            first col and last col+1 to be read, starting from 0

        Returns
        -------
        data: pandas.DataFrame, pandas.Series or float
            depending on the shape of the requested data

        """
        # read data
        data = Excels.read(
            self.file,
            self.tab)[rows[0]:rows[1], cols[0]:cols[1]].copy()

        shape = data.shape

        # empty cells
        if shape[0] == 0 or shape[1] == 0:
            raise ValueError(
                self.py_name + "\n"
                "The cells are empty.\n"
                + self._file_sheet
            )

        # if it is a single row remove its dimension
        if shape[1] == 1:
            data = data[:, 0]
        if shape[0] == 1:
            data = data[0]
        return data

    def _get_data_from_file_opyxl(self, cellname):
        """
        Function to read data from excel file using cell range name

        Parameters
        ----------
        cellname: str
            the cell range name

        Returns
        -------
        data: numpy.ndarray or float
            depending on the shape of the requested data
        shape: list
            The shape of the data in 2D.

        """
        # read data
        try:
            excel = Excels.read_opyxl(self.file)
        except InvalidFileException:
            raise ValueError(
                self.py_name + "\n"
                f"Cannot read the file '{self.file}'...\n"
                f"It could happen that cell='{cellname}' was "
                "read as a cell range name due to a wrong "
                "definition of cell value"
            )

        # Get global and local cellrange names
        global_cellranges = excel.defined_names
        local_cellranges = None
        # need to lower the sheetnames as Vensim has no case sensitivity
        for sheet in excel.sheetnames:
            if sheet.lower() == self.tab.lower():
                local_cellranges = excel[sheet].defined_names
                break

        if local_cellranges is None:
            # Error if it is not able to get the localSheetId
            raise ValueError(
                self.py_name + "\n"
                "The sheet doesn't exist...\n"
                + self._file_sheet
            )
        try:
            # Search for local and global names
            cellrange = local_cellranges.get(cellname)\
                        or global_cellranges.get(cellname)
            sheet, cells = next(cellrange.destinations)

            assert sheet.lower() == self.tab.lower()
            self.tab = sheet  # case insensitivity in sheet name

            # Get the cells where the cellrange is defined
            cells = re.split(r":|\$", cells)
            cols = [self._col_to_num(cells[1]), None]
            rows = [int(cells[2])-1, None]
            if len(cells) == 3:
                # 0 dim cell range
                cols[1] = cols[0]+1
                rows[1] = rows[0]+1
            else:
                # array or table
                cols[1] = self._col_to_num(cells[4])+1
                rows[1] = int(cells[5])
            # Use pandas to read the data and return its original shape
            return self._get_data_from_file(rows, cols), \
                [rows[1]-rows[0], cols[1]-cols[0]]

        except (AttributeError, AssertionError):
            # key error if the cellrange doesn't exist in the file or sheet
            raise AttributeError(
               self.py_name + "\n"
               f"The cellrange name '{cellname}'\n"
               "Doesn't exist in:\n" + self._file_sheet
               )

    def _get_series_data(self, series_across, series_row_or_col, cell, size):
        """
        Function thar reads series and data from excel file for
        DATA and LOOKUPS.

        Parameters
        ----------
        series_across: "row", "column" or "name"
            The way to read series file.
        series_row_or_col: int or str
            If series_across is "row" the row number where the series data is.
            If series_across is "column" the column name where the series
            data is.
            If series_across is "name" the cell range name where the series
            data is.
        cell:
            If series_across is not "name, the top left cell where the
            data table starts.
            Else the name of the cell range where the data is.
        size:
            The size of the 2nd dimension of the data.

        Returns
        -------
        series, data: ndarray (1D), ndarray(1D/2D)
            The values of the series and data.

        """
        if series_across == "row":
            # Horizontal data (dimension values in a row)

            # get the dimension values
            first_row, first_col = self._split_excel_cell(cell)
            series = self._get_data_from_file(
                rows=[int(series_row_or_col)-1, int(series_row_or_col)],
                cols=[first_col, None])

            # read data
            data = self._get_data_from_file(
                rows=[first_row, first_row + size],
                cols=[first_col, None]).transpose()

        elif series_across == "column":
            # Vertical data (dimension values in a column)

            # get the dimension values
            first_row, first_col = self._split_excel_cell(cell)
            series_col = self._col_to_num(series_row_or_col)
            series = self._get_data_from_file(
                rows=[first_row, None],
                cols=[series_col, series_col+1])

            # read data
            data = self._get_data_from_file(
                rows=[first_row, None],
                cols=[first_col, first_col + size])

        else:
            # get series data
            series, s_shape = self._get_data_from_file_opyxl(series_row_or_col)

            if isinstance(series, float):
                series = np.array([series])

            if s_shape[0] > 1 and s_shape[1] > 1:
                # Error if the lookup/time dimension is 2D
                raise ValueError(
                  self.py_name + "\n"
                  + "Dimension given in:\n"
                  + self._file_sheet
                  + "\tDimension name:"
                  + "\t'{}'\n".format(series_row_or_col)
                  + " is a table and not a vector"
                  )
            elif s_shape[1] != 1:
                transpose = True
            else:
                transpose = False

            # get data
            data, d_shape = self._get_data_from_file_opyxl(cell)

            if isinstance(data, float):
                data = np.array([data])

            if transpose:
                # transpose for horizontal definition of dimension
                data = data.transpose()
                d_shape = d_shape[1], d_shape[0]

            if d_shape[0] != len(series):
                raise ValueError(
                  self.py_name + "\n"
                  + "Dimension and data given in:\n"
                  + self._file_sheet
                  + "\tDimension name:\t'{}'\n".format(series_row_or_col)
                  + "\tData name:\t'{}'\n".format(cell)
                  + " don't have the same length in the 1st dimension"
                  )

            if d_shape[1] != size:
                # Given coordinates length is different than
                # the lentgh of 2nd dimension
                raise ValueError(
                  self.py_name + "\n"
                  + "Data given in:\n"
                  + self._file_sheet
                  + "\tData name:\t'{}'\n".format(cell)
                  + " has not the same size as the given coordinates"
                  )

        return series, data

    def _resolve_file(self, root):
        """
        Resolve input file path. Joining the file with the root and
        checking if it exists.

        Parameters
        ----------
        root: pathlib.Path or str
            The root path to the model file.

        Returns
        -------
        None

        """
        if str(self.file)[0] == '?':
            # TODO add an option to include indirect references
            raise ValueError(
                self.py_name + "\n"
                + f"Indirect reference to file: '{self.file}'")

        # Join path and resolve it to better print error messages
        self.file = root.joinpath(self.file).resolve()

        if not self.file.is_file():
            raise FileNotFoundError(
                self.py_name + "\n"
                + "File '%s' not found." % self.file)

    def _initialize_data(self, element_type):
        """
        Initialize one element of DATA or LOOKUPS

        Parameters
        ----------
        element_type: str
            "lookup" for LOOKUPS, "data" for data.

        Returns
        -------
        data: xarray.DataArray
            Dataarray with the time or interpolation dimension
            as first dimension.

        """
        self._resolve_file(root=self.root)
        series_across = self._series_selector(self.x_row_or_col, self.cell)
        size = utils.compute_shape(self.coords, reshape_len=1,
                                   py_name=self.py_name)[0]

        series, data = self._get_series_data(
            series_across=series_across,
            series_row_or_col=self.x_row_or_col,
            cell=self.cell, size=size
        )

        # remove nan or missing values from dimension
        if series_across != "name":
            # Remove last nans only if the method is to read by row or col
            i = 0
            try:
                while np.isnan(series[i-1]):
                    i -= 1
            except IndexError:
                # series has len 0
                raise ValueError(
                    self.py_name + "\n"
                    + "Dimension given in:\n"
                    + self._file_sheet
                    + "\t{}:\t'{}'\n".format(series_across, self.x_row_or_col)
                    + " has length 0"
                    )

            if i != 0:
                series = series[:i]
                data = data[:i]

        # warning/error if missing data in the series
        if any(np.isnan(series)) and self.missing != "keep":
            valid_values = ~np.isnan(series)
            series = series[valid_values]
            data = data[valid_values]
            if all(np.isnan(series)):
                raise ValueError(
                    self.py_name + "\n"
                    + "Dimension given in:\n"
                    + self._file_sheet
                    + "\t{}:\t'{}'\n".format(series_across, self.x_row_or_col)
                    + " has length 0"
                    )
            if self.missing == "warning":
                warnings.warn(
                  self.py_name + "\n"
                  + "Dimension value missing or non-valid in:\n"
                  + self._file_sheet
                  + "\t{}:\t'{}'\n".format(series_across, self.x_row_or_col)
                  + " the corresponding data value(s) to the "
                  + "missing/non-valid value(s) will be ignored\n\n"
                  )
            elif self.missing == "raise":
                raise ValueError(
                  self.py_name + "\n"
                  + "Dimension value missing or non-valid in:\n"
                  + self._file_sheet
                  + "\t{}:\t'{}'\n".format(series_across, self.x_row_or_col)
                  )

        # reorder data with increasing series
        if not np.all(np.diff(series) > 0) and self.missing != "keep":
            order = np.argsort(series)
            series = series[order]
            data = data[order]
            # Check if the lookup/time dimension is well defined
            if np.any(np.diff(series) == 0):
                raise ValueError(self.py_name + "\n"
                                 + "Dimension given in:\n"
                                 + self._file_sheet
                                 + "\t{}:\t'{}'\n".format(
                                    series_across, self.x_row_or_col)
                                 + " has repeated values")

        # Check for missing values in data
        if np.any(np.isnan(data)) and self.missing != "keep":
            if series_across == "name":
                cell_type = "Cellrange"
            else:
                cell_type = "Reference cell"

            if self.missing == "warning":
                # Fill missing values with the chosen interpolation method
                # what Vensim does during running for DATA
                if self.interp != "raw":
                    interpolate_message =\
                        " the corresponding value will be filled "\
                        + "with the interpolation method of the object."
                else:
                    interpolate_message = ""

                warnings.warn(
                    self.py_name + "\n"
                    + "Data value missing or non-valid in:\n"
                    + self._file_sheet
                    + "\t{}:\t'{}'\n".format(cell_type, self.cell)
                    + interpolate_message + "\n\n"
                    )
            elif self.missing == "raise":
                raise ValueError(
                    self.py_name + "\n"
                    + "Data value missing or non-valid in:\n"
                    + self._file_sheet
                    + "\t{}:\t'{}'\n".format(cell_type, self.cell)
                    )
            # fill values
            if self.interp != "raw":
                self._fill_missing(series, data)

        reshape_dims = tuple([len(series)] + utils.compute_shape(self.coords))

        if len(reshape_dims) > 1:
            data = self._reshape(data, reshape_dims)

        if element_type == "lookup":
            dim_name = "lookup_dim"
        else:
            dim_name = "time"

        data = xr.DataArray(
            data=data,
            coords={dim_name: series, **self.coords},
            dims=[dim_name] + list(self.coords)
        )

        return data

    def _fill_missing(self, series, data):
        """
        Fills missing values in excel read data. Mutates the values in data.

        Parameters
        ----------
        series:
          the time series without missing values
        data:
          the data with missing values

        Returns
        -------
        None
        """
        # if data is 2dims we need to interpolate
        datanan = np.isnan(data)
        keeping_nan = False
        if len(data.shape) == 1:
            if not np.all(datanan):
                data[datanan] = self._interpolate_missing(
                    series[datanan],
                    series[~datanan],
                    data[~datanan])
            else:
                keeping_nan = True
        else:
            for i, nanlist in enumerate(list(datanan.transpose())):
                if not np.all(nanlist):
                    data[nanlist, i] = self._interpolate_missing(
                        series[nanlist],
                        series[~nanlist],
                        data[~nanlist][:, i])
                else:
                    keeping_nan = True

        if keeping_nan:
            warnings.warn(
                "Not able to interpolate some values..."
                " keeping them as missing.\n")

    def _interpolate_missing(self, x, xr, yr):
        """
        Interpolates a list of missing values from _fill_missing

        Parameters
        ----------
        x:
          list of missing values interpolate
        xr:
          non-missing x values
        yr:
          non-missing y values

        Returns
        -------
        y:
          Result after interpolating x with self.interp method

        """
        y = np.empty_like(x, dtype=float)
        for i, value in enumerate(x):
            if value >= xr[-1]:
                y[i] = yr[-1]
            elif value <= xr[0]:
                y[i] = yr[0]
            elif self.interp == 'look_forward':
                y[i] = yr[xr >= value][0]
            elif self.interp == 'hold_backward':
                y[i] = yr[xr <= value][-1]
            else:
                y[i] = np.interp(value, xr, yr)
        return y

    @property
    def _file_sheet(self):
        """
        Returns file and sheet name in a string
        """
        return "\tFile name:\t'{}'\n".format(self.file)\
               + "\tSheet name:\t'{}'\n".format(self.tab)

    @staticmethod
    def _col_to_num(col):
        """
        Transforms the column name to int

        Parameters
        ----------
        col: str
          Column name

        Returns
        -------
        int
          Column number
        """
        if len(col) == 1:
            return ord(col.upper()) - ord('A')
        elif len(col) == 2:
            left = ord(col[0].upper()) - ord('A') + 1
            right = ord(col[1].upper()) - ord('A')
            return left * (ord('Z')-ord('A')+1) + right
        else:
            left = ord(col[0].upper()) - ord('A') + 1
            center = ord(col[1].upper()) - ord('A') + 1
            right = ord(col[2].upper()) - ord('A')
            return left * ((ord('Z')-ord('A')+1)**2)\
                + center * (ord('Z')-ord('A')+1)\
                + right

    def _split_excel_cell(self, cell):
        """
        Splits a cell value given in a string.
        Returns None for non-valid cell formats.

        Parameters
        ----------
        cell: str
          Cell like string, such as "A1", "b16", "AC19"...
          If it is not a cell like string will return None.

        Returns
        -------
        row number, column number: int, int
          If the cell input is valid. Both numbers are given in Python
          enumeration, i.e., first row and first column are 0.

        """
        split = re.findall(r'\d+|\D+', cell)
        try:
            # check that we only have two values [column, row]
            assert len(split) == 2
            # check that the column name has no special characters
            assert not re.compile('[^a-zA-Z]+').search(split[0])
            # check that row number is not 0
            assert int(split[1]) != 0
            # the column name has as maximum 3 letters
            assert len(split[0]) <= 3
            return int(split[1])-1, self._col_to_num(split[0])
        except AssertionError:
            return

    @staticmethod
    def _reshape(data, dims):
        """
        Reshapes an pandas.DataFrame, pandas.Series, xarray.DataArray
        or np.ndarray in the given dimensions.

        Parameters
        ----------
        data: xarray.DataArray/numpy.ndarray
          Data to be reshaped
        dims: tuple
          The dimensions to reshape.

        Returns
        -------
        numpy.ndarray
          reshaped array
        """
        if isinstance(data, (float, int)):
            data = np.array(data)
        elif isinstance(data, xr.DataArray):
            data = data.values

        return data.reshape(dims)

    def _series_selector(self, x_row_or_col, cell):
        """
        Selects if a series data (DATA/LOOKUPS), should be read by columns,
        rows or cellrange name.
        Based on the input format of x_row_or_col and cell.
        The format of the 2 variables must be consistent.

        Parameters
        ----------
        x_row_or_col: str
          String of a number if series is given in a row, letter if series is
          given in a column or name if the series is given by cellrange name.
        cell: str
          Cell identificator, such as "A1", or name if the data is given
          by cellrange name.

        Returns
        -------
        series_across: str
          "row" if series is given in a row
          "column" if series is given in a column
          "name" if series and data are given by range name

        """
        try:
            # if x_row_or_col is numeric the series must be a row
            int(x_row_or_col)
            return "row"

        except ValueError:
            if self._split_excel_cell(cell):
                # if the cell can be splitted means that the format is
                # "A1" like then the series must be a column
                return "column"
            else:
                return "name"


class ExtData(External, Data):
    """
    Class for Vensim GET XLS DATA/GET DIRECT DATA
    """

    def __init__(self, file_name, tab, time_row_or_col, cell,
                 interp, coords, root, final_coords, py_name):
        super().__init__(py_name)
        self.files = [file_name]
        self.tabs = [tab]
        self.time_row_or_cols = [time_row_or_col]
        self.cells = [cell]
        self.coordss = [coords]
        self.root = root
        self.final_coords = final_coords
        self.interp = interp or "interpolate"
        self.is_float = not bool(coords)

        # check if the interpolation method is valid
        if self.interp not in ["interpolate", "raw",
                               "look_forward", "hold_backward"]:
            raise ValueError(self.py_name + "\n"
                             + " The interpolation method (interp) must be "
                             + "'raw', 'interpolate', "
                             + "'look_forward' or 'hold_backward'")

    def add(self, file_name, tab, time_row_or_col, cell, interp, coords):
        """
        Add information to retrieve new dimension in an already declared object
        """
        self.files.append(file_name)
        self.tabs.append(tab)
        self.time_row_or_cols.append(time_row_or_col)
        self.cells.append(cell)
        self.coordss.append(coords)

        interp = interp or "interpolate"

        if interp.replace(" ", "_") != self.interp:
            raise ValueError(self.py_name + "\n"
                             + "Error matching interpolation method with "
                             + "previously defined one")

        if list(coords) != list(self.coordss[0]):
            raise ValueError(self.py_name + "\n"
                             + "Error matching dimensions with previous data")

    def initialize(self):
        """
        Initialize all elements and create the self.data xarray.DataArray
        """
        if not self.coordss[0]:
            # Just load one value (no add)
            for self.file, self.tab, self.x_row_or_col, \
                self.cell, self.coords\
                in zip(self.files, self.tabs, self.time_row_or_cols,
                       self.cells, self.coordss):
                self.data = self._initialize_data("data")
        else:
            # Load in several lines (add)
            self.data = xr.DataArray(
                np.nan, self.final_coords, list(self.final_coords))

            for self.file, self.tab, self.x_row_or_col, \
                self.cell, self.coords\
                in zip(self.files, self.tabs, self.time_row_or_cols,
                       self.cells, self.coordss):
                values = self._initialize_data("data")

                coords = {"time": values.coords["time"].values, **self.coords}
                if "time" not in self.data.dims:
                    self.data = self.data.expand_dims(
                        {"time": coords["time"]}, axis=0).copy()

                self.data.loc[coords] = values.values

        # set what to return when raw
        if self.final_coords:
            self.nan = xr.DataArray(
                np.nan, self.final_coords, list(self.final_coords))
        else:
            self.nan = np.nan


class ExtLookup(External, Lookups):
    """
    Class for Vensim GET XLS LOOKUPS/GET DIRECT LOOKUPS
    """

    def __init__(self, file_name, tab, x_row_or_col, cell, coords,
                 root, final_coords, py_name):
        super().__init__(py_name)
        self.files = [file_name]
        self.tabs = [tab]
        self.x_row_or_cols = [x_row_or_col]
        self.cells = [cell]
        self.coordss = [coords]
        self.root = root
        self.final_coords = final_coords
        self.interp = "interpolate"
        self.is_float = not bool(coords)

    def add(self, file_name, tab, x_row_or_col, cell, coords):
        """
        Add information to retrieve new dimension in an already declared object
        """
        self.files.append(file_name)
        self.tabs.append(tab)
        self.x_row_or_cols.append(x_row_or_col)
        self.cells.append(cell)
        self.coordss.append(coords)

        if list(coords) != list(self.coordss[0]):
            raise ValueError(self.py_name + "\n"
                             + "Error matching dimensions with previous data")

    def initialize(self):
        """
        Initialize all elements and create the self.data xarray.DataArray
        """
        if not self.coordss[0]:
            # Just loag one value (no add)
            for self.file, self.tab, self.x_row_or_col, \
                self.cell, self.coords\
                in zip(self.files, self.tabs, self.x_row_or_cols,
                       self.cells, self.coordss):
                self.data = self._initialize_data("lookup")
        else:
            # Load in several lines (add)
            self.data = xr.DataArray(
                np.nan, self.final_coords, list(self.final_coords))

            for self.file, self.tab, self.x_row_or_col, \
                self.cell, self.coords\
                in zip(self.files, self.tabs, self.x_row_or_cols,
                       self.cells, self.coordss):
                values = self._initialize_data("lookup")

                coords = {
                    "lookup_dim": values.coords["lookup_dim"].values,
                    **self.coords
                }
                if "lookup_dim" not in self.data.dims:
                    self.data = self.data.expand_dims(
                        {"lookup_dim": coords["lookup_dim"]}, axis=0).copy()

                self.data.loc[coords] = values.values


class ExtConstant(External):
    """
    Class for Vensim GET XLS CONSTANTS/GET DIRECT CONSTANTS
    """

    def __init__(self, file_name, tab, cell, coords,
                 root, final_coords, py_name):
        super().__init__(py_name)
        self.files = [file_name]
        self.tabs = [tab]
        self.transposes = [
            cell[-1] == '*' and np.prod(utils.compute_shape(coords)) > 1]
        self.cells = [cell.strip('*')]
        self.coordss = [coords]
        self.root = root
        self.final_coords = final_coords

    def add(self, file_name, tab, cell, coords):
        """
        Add information to retrieve new dimension in an already declared object
        """
        self.files.append(file_name)
        self.tabs.append(tab)
        self.transposes.append(
            cell[-1] == '*' and np.prod(utils.compute_shape(coords)) > 1)
        self.cells.append(cell.strip('*'))
        self.coordss.append(coords)

        if list(coords) != list(self.coordss[0]):
            raise ValueError(self.py_name + "\n"
                             + "Error matching dimensions with previous data")

    def initialize(self):
        """
        Initialize all elements and create the self.data xarray.DataArray
        """
        if not self.coordss[0]:
            # Just loag one value (no add)
            for self.file, self.tab, self.transpose, self.cell, self.coords\
                in zip(self.files, self.tabs, self.transposes,
                       self.cells, self.coordss):
                self.data = self._initialize()
        else:
            # Load in several lines (add)

            self.data = xr.DataArray(
                np.nan, self.final_coords, list(self.final_coords))

            for self.file, self.tab, self.transpose, self.cell, self.coords\
                in zip(self.files, self.tabs, self.transposes,
                       self.cells, self.coordss):
                self.data.loc[self.coords] = self._initialize().values

    def _initialize(self):
        """
        Initialize one element
        """
        self._resolve_file(root=self.root)
        split = self._split_excel_cell(self.cell)
        if split:
            data_across = "cell"
            cell = split
        else:
            data_across = "name"
            cell = self.cell

        shape = utils.compute_shape(self.coords, reshape_len=2,
                                    py_name=self.py_name)

        if self.transpose:
            shape.reverse()

        data = self._get_constant_data(data_across, cell, shape)

        if self.transpose:
            data = data.transpose()

        if np.any(np.isnan(data)):
            # nan values in data
            if data_across == "name":
                cell_type = "Cellrange"
            else:
                cell_type = "Reference cell"

            if self.missing == "warning":
                warnings.warn(
                    self.py_name + "\n"
                    + "Constant value missing or non-valid in:\n"
                    + self._file_sheet
                    + "\t{}:\t'{}'\n".format(cell_type, self.cell)
                    )
            elif self.missing == "raise":
                raise ValueError(
                    self.py_name + "\n"
                    + "Constant value missing or non-valid in:\n"
                    + self._file_sheet
                    + "\t{}:\t'{}'\n".format(cell_type, self.cell)
                    )

        # Create only an xarray if the data is not 0 dimensional
        if len(self.coords) > 0:
            reshape_dims = tuple(utils.compute_shape(self.coords))

            if len(reshape_dims) > 1:
                data = self._reshape(data, reshape_dims)

            data = xr.DataArray(
                data=data, coords=self.coords, dims=list(self.coords)
            )

        return data

    def _get_constant_data(self, data_across, cell, shape):
        """
        Function thar reads data from excel file for CONSTANT

        Parameters
        ----------
        data_across: "cell" or "name"
            The way to read data file.
        cell: int or str
            If data_across is "cell" the lefttop split cell value where
            the data is.
            If data_across is "name" the cell range name where the data is.
        shape: list
            The shape of the data in 2D.

        Returns
        -------
        data: float/ndarray(1D/2D)
            The values of the data.

        """
        if data_across == "cell":
            # read data from topleft cell name using pandas
            start_row, start_col = cell

            return self._get_data_from_file(
                rows=[start_row, start_row + shape[0]],
                cols=[start_col, start_col + shape[1]])

        else:
            # read data from cell range name using OpenPyXL
            data, xl_shape = self._get_data_from_file_opyxl(cell)
            if shape != xl_shape:
                raise ValueError(self.py_name + "\n"
                                 + "Data given in:\n"
                                 + self._file_sheet
                                 + "\tData name:\t{}\n".format(cell)
                                 + " has not the same shape as the"
                                 + " given coordinates")

            return data

    def __call__(self):
        return self.data


class ExtSubscript(External):
    """
    Class for Vensim GET XLS SUBSCRIPT/GET DIRECT SUBSCRIPT
    """

    def __init__(self, file_name, tab, firstcell, lastcell, prefix, root):
        super().__init__("Hardcoded external subscript")
        self.file = file_name
        self.tab = tab
        self.prefix = prefix
        self._resolve_file(root=root)
        split = self._split_excel_cell(firstcell)
        if split:
            subs = self.get_subscripts_cell(*split, lastcell)
        else:
            subs = self.get_subscripts_name(firstcell)

        self.subscript = [
            self.prefix + str(d) for d in subs.flatten()
            if self._not_nan(d)
            ]

    def get_subscripts_cell(self, row_first, col_first, lastcell):
        """Get subscripts from common cell definition"""
        if not lastcell:
            row_last, col_last = None, None
        else:
            split = self._split_excel_cell(lastcell)
            if split:
                # last cell is col and row
                row_last, col_last = split
            elif lastcell.isdigit():
                # last cell is row number only
                row_last = int(lastcell)-1
                col_last = None
            else:
                # last cell is a col value only
                row_last = None
                col_last = self._col_to_num(lastcell)

        # update read keywargs for rows and columns to read
        read_kwargs = {}
        if row_last is not None:
            read_kwargs['nrows'] = row_last-row_first+1
        if col_last is not None:
            read_kwargs['usecols'] = np.arange(col_first, col_last+1)

        # get the function to read the data based on its extension
        ext = self.file.suffix.lower()
        if ext in _SPREADSHEET_EXTS:
            read_func = pd.read_excel
            read_kwargs['sheet_name'] = self.tab
        elif ext == '.csv':
            read_func = pd.read_csv
            if self.tab and not self.tab[0].isalnum():
                read_kwargs['sep'] = self.tab
        else:
            read_func = pd.read_table
            if self.tab and not self.tab[0].isalnum():
                read_kwargs['sep'] = self.tab

        # read the data
        data = read_func(
            self.file,
            skiprows=row_first,
            dtype=object,
            header=None,
            **read_kwargs
            ).values

        # skip columns if usecols couldn't be used
        if col_last is None:
            data = data[:, col_first:]

        return data

    def get_subscripts_name(self, cellname):
        """Get subscripts from cell range name definition"""
        try:
            excel = load_workbook(self.file, read_only=True, data_only=True)
        except InvalidFileException:
            raise ValueError(
                self.py_name + "\n"
                f"Cannot read the file '{self.file}'...\n"
                f"It could happen that firstcell='{cellname}' was "
                "read as a cell range name due to a wrong definition "
                "of cell value"
            )
        global_cellranges = excel.defined_names
        local_cellranges = None
        # need to lower the sheetnames as Vensim has no case sensitivity
        for sheet in excel.sheetnames:
            if sheet.lower() == self.tab.lower():
                local_cellranges = excel[sheet].defined_names
                break

        if local_cellranges is None:
            # Error if it is not able to get the localSheetId
            raise ValueError(
                self.py_name + "\n"
                "The sheet doesn't exist...\n"
                + self._file_sheet
            )
        try:
            # Search for local and global names
            cellrange = local_cellranges.get(cellname)\
                        or global_cellranges.get(cellname)
            sheet, cells = next(cellrange.destinations)

            assert sheet.lower() == self.tab.lower()
            self.tab = sheet  # case insensitivity in sheet name

            # Get the cells where the cellrange is defined
            first_cell, last_cell = cells.replace("$", '').split(":")
        except (AttributeError, AssertionError):
            # key error if the cellrange doesn't exist in the file or sheet
            raise AttributeError(
               self.py_name + "\n"
               f"The cellrange name '{cellname}'\n"
               "Doesn't exist in:\n" + self._file_sheet
               )
        else:
            return self.get_subscripts_cell(
                *self._split_excel_cell(first_cell), last_cell)

    @staticmethod
    def _not_nan(value):
        """Check if a value is not nan"""
        if isinstance(value, str):
            return True
        return not np.isnan(value)



"""
The provided functions have no direct analog in the standard Python data
analytics stack, or require information about the internal state of the
system beyond what is present in the function call. They are provided
in a structure that makes it easy for the model elements to call. The
functions may be similar to the original functions given by Vensim or
Stella, but sometimes the number or order of arguments may change.
"""
import warnings
from datetime import datetime

import numpy as np
import xarray as xr

from . import utils

SMALL_VENSIM = 1e-6  # What is considered zero according to Vensim Help


def ramp(time, slope, start, finish=None):
    """
    Implements vensim's and xmile's RAMP function.

    Parameters
    ----------
    time: pysd.py_backend.components.Time
        Model time object.
    slope: float or xarray.DataArray
        The slope of the ramp starting at zero at time start.
    start: float or xarray.DataArray
        Time at which the ramp begins.
    finish: float or xarray.DataArray or None (oprional)
        Time at which the ramp ends. If None the ramp will never end.
        Default is None.

    Returns
    -------
    float or xarray.DataArray:
        If prior to ramp start, returns zero.
        If after ramp ends, returns top of ramp.

    """
    t = time()
    # compute out (MIN(time, finish)-start)*slope
    if finish is None:
        final = t
    else:
        final = np.minimum(finish, t)

    return (t + SMALL_VENSIM > start) * slope * (final-start)


def step(time, value, tstep):
    """"
    Implements vensim's STEP function.

    Parameters
    ----------
    time: pysd.py_backend.components.Time
        Model time object.
    value: float or xarray.DataArray
        The height of the step.
    tstep: float or xarray.DataArray
        The time at and after which `result` equals `value`.

    Returns
    -------
    float or xarray.DataArray:
        - In range [-inf, tstep):
            returns 0
        - In range [tstep, +inf]:
            returns `value`
    """
    return value * (time() + time.time_step()/2 > tstep)


def pulse(time, start, repeat_time=0, width=None, magnitude=None, end=None):
    """
    Implements Vensim's PULSE and PULSE TRAIN functions and Xmile's PULSE
    function.

    Parameters
    ----------
    time: pysd.py_backend.components.Time
        Model time object.
    start: float
        Starting time of the pulse.
    repeat_time: float (optional)
        Time interval of the pulse repetition. If 0 it will return a
        single pulse. Default is 0.
    width: float or None (optional)
        Duration of the pulse. If None only one-time_step pulse will be
        generated. Default is None.
    magnitude: float or None (optional)
        The magnitude of the pulse. If None it will return 1 when the
        pulse happens, similar to magnitude=time_step(). Default is None.
    end: float or None (optional)
        Final time of the pulse. If None there is no final time.
        Default is None.

    Returns
    -------
    float or xarray.DataArray:
        - In range [-inf, start):
            returns 0
        - In range [start + n*repeat_time, start + n*repeat_time + width):
            returns magnitude/time_step or 1
        - In range [start + n*repeat_time + width, start + (n+1)*repeat_time):
            returns 0

    """
    t = time()
    width = .5*time.time_step() if width is None else width
    out = magnitude/time.time_step() if magnitude is not None else 1
    if repeat_time == 0:
        return out if start - SMALL_VENSIM <= t < start + width else 0
    elif start <= t and (end is None or t < end):
        return out if (t - start + SMALL_VENSIM) % repeat_time < width else 0
    else:
        return 0


def if_then_else(condition, val_if_true, val_if_false):
    """
    Implements Vensim's IF THEN ELSE function.
    https://www.vensim.com/documentation/20475.htm

    Parameters
    ----------
    condition: bool or xarray.DataArray of bools
    val_if_true: callable
        Value to evaluate and return when condition is true.
    val_if_false: callable
        Value to evaluate and return when condition is false.

    Returns
    -------
    float or xarray.DataArray:
        The value depending on the condition.

    """
    # NUMPY: replace xr by np
    if isinstance(condition, xr.DataArray):
        # NUMPY: neccessarry for keep the same shape always
        # if condition.all():
        #    value = val_if_true()
        # elif not condition.any():
        #    value = val_if_false()
        # else:
        #    return np.where(condition, val_if_true(), val_if_false())
        #
        # if isinstance(value, np.ndarray):
        #    return value
        # return np.full_like(condition, value)
        if condition.all():
            return val_if_true()
        elif not condition.any():
            return val_if_false()

        return xr.where(condition, val_if_true(), val_if_false())

    return val_if_true() if condition else val_if_false()


def xidz(numerator, denominator, x):
    """
    Implements Vensim's XIDZ function.
    https://www.vensim.com/documentation/fn_xidz.htm

    This function executes a division, robust to denominator being zero.
    In the case of zero denominator, the final argument is returned.

    Parameters
    ----------
    numerator: float or xarray.DataArray
        Numerator of the operation.
    denominator: float or xarray.DataArray
        Denominator of the operation.
    x: float or xarray.DataArray
        The value to return if the denominator is zero.

    Returns
    -------
    float or xarray.DataArray:
        - numerator/denominator if denominator > small_vensim
        - value_if_denom_is_zero otherwise

    """
    # NUMPY: replace DataArray by np.ndarray, xr.where -> np.where
    if isinstance(denominator, xr.DataArray):
        return xr.where(np.abs(denominator) < SMALL_VENSIM,
                        x,
                        numerator/denominator)

    if abs(denominator) < SMALL_VENSIM:
        # NUMPY: neccessarry for keep the same shape always
        # if isinstance(numerator, np.ndarray):
        #    return np.full_like(numerator, x)
        return x
    else:
        # NUMPY: neccessarry for keep the same shape always
        # if isinstance(x, np.ndarray):
        #    return np.full_like(x, numerator/denominator)
        return numerator/denominator


def zidz(numerator, denominator):
    """
    This function bypasses divide-by-zero errors,
    implementing Vensim's ZIDZ function.
    https://www.vensim.com/documentation/fn_zidz.htm

    Parameters
    ----------
    numerator: float or xarray.DataArray
        value to be divided
    denominator: float or xarray.DataArray
        value to devide by

    Returns
    -------
    float or xarray.DataArray:
        - numerator/denominator if denominator > small_vensim
        - 0 or 0s array otherwise

    """
    # NUMPY: replace DataArray by np.ndarray, xr.where -> np.where
    if isinstance(denominator, xr.DataArray):
        return xr.where(np.abs(denominator) < SMALL_VENSIM,
                        0,
                        numerator/denominator)

    if abs(denominator) < SMALL_VENSIM:
        # NUMPY: neccessarry for keep the same shape always
        # if isinstance(denominator, np.ndarray):
        #    return np.zeros_like(denominator)
        if isinstance(numerator, xr.DataArray):
            return xr.DataArray(0, numerator.coords, numerator.dims)
        return 0
    else:
        return numerator/denominator


def active_initial(stage, expr, init_val):
    """
    Implements vensim's ACTIVE INITIAL function

    Parameters
    ----------
    stage: str
        The stage of the model.
    expr: callable
        Running stage value
    init_val: float or xarray.DataArray
        Initialization stage value.

    Returns
    -------
    float or xarray.DataArray:
        - inti_val if stage='Initialization'
        - expr() otherwise
    """
    # NUMPY: both must have same dimensions in inputs, remove time.stage
    if stage == 'Initialization':
        return init_val
    else:
        return expr()


def incomplete(*args):
    """
    Implements an incomplete functions.
    Prompts a RuntimeWarning.

    Parameters
    ----------
    *args: arguments

    Returns
    -------
    numpy.nan

    """
    warnings.warn(
        'Call to undefined function, calling dependencies and returning NaN',
        RuntimeWarning, stacklevel=2)

    return np.nan


def not_implemented_function(*args):
    """
    Implements a not implemented functions.
    Raises a NotImplementedError if it is called.

    Parameters
    ----------
    *args: arguments
        The first argument must be the name of the function as str to
        properly print the error message.

    """
    raise NotImplementedError(f"Not implemented function '{args[0]}'")


def integer(x):
    """
    Implements Vensim's INTEGER function.

    Parameters
    ----------
    x: float or xarray.DataArray
        Input value.

    Returns
    -------
    integer: float or xarray.DataArray
        Returns integer part of x.

    """
    # NUMPY: replace xr by np
    if isinstance(x, xr.DataArray):
        return x.astype(int)
    else:
        return int(x)


def quantum(a, b):
    """
    Implements Vensim's QUANTUM function.

    Parameters
    ----------
    a: float or xarray.DataArray
        Input value.
    b: float or xarray.DataArray
        Input value.

    Returns
    -------
    quantum: float or xarray.DataArray
        If b > 0 returns b * integer(a/b). Otherwise, returns a.

    """
    # NUMPY: replace xr by np
    if isinstance(b, xr.DataArray):
        return xr.where(b < SMALL_VENSIM, a, b*integer(a/b))
    if b < SMALL_VENSIM:
        return a
    else:
        return b*integer(a/b)


def modulo(x, m):
    """
    Implements Vensim's MODULO function.

    Parameters
    ----------
    x: float or xarray.DataArray
        Input value.
    m: float or xarray.DataArray
        Modulo to compute.

    Returns
    -------
    modulo: float or xarray.DataArray
        Returns x modulo m, if x is smaller than 0 the result is given
        in the range (-m, 0] as Vensim does. x - quantum(x, m)

    """
    return x - quantum(x, m)


def sum(x, dim=None):
    """
    Implements Vensim's SUM function.

    Parameters
    ----------
    x: xarray.DataArray
        Input value.
    dim: list of strs (optional)
        Dimensions to apply the function over.
        If not given the function will be applied over all dimensions.

    Returns
    -------
    sum: xarray.DataArray or float
        The result of the sum operation in the given dimensions.

    """
    # NUMPY: replace by np.sum(x, axis=axis) put directly in the file
    # float returned if the function is applied over all the dimensions
    if dim is None or set(x.dims) == set(dim):
        return float(x.sum())

    return x.sum(dim=dim)


def prod(x, dim=None):
    """
    Implements Vensim's PROD function.

    Parameters
    ----------
    x: xarray.DataArray
        Input value.
    dim: list of strs (optional)
        Dimensions to apply the function over.
        If not given the function will be applied over all dimensions.

    Returns
    -------
    prod: xarray.DataArray or float
        The result of the product operation in the given dimensions.

    """
    # NUMPY: replace by np.prod(x, axis=axis) put directly in the file
    # float returned if the function is applied over all the dimensions
    if dim is None or set(x.dims) == set(dim):
        return float(x.prod())

    return x.prod(dim=dim)


def vmin(x, dim=None):
    """
    Implements Vensim's Vmin function.

    Parameters
    ----------
    x: xarray.DataArray
        Input value.
    dim: list of strs (optional)
        Dimensions to apply the function over.
        If not given the function will be applied over all dimensions.

    Returns
    -------
    vmin: xarray.DataArray or float
        The result of the minimum value over the given dimensions.

    """
    # NUMPY: replace by np.min(x, axis=axis) put directly in the file
    # float returned if the function is applied over all the dimensions
    if dim is None or set(x.dims) == set(dim):
        return float(x.min())

    return x.min(dim=dim)


def vmax(x, dim=None):
    """
    Implements Vensim's VMAX function.

    Parameters
    ----------
    x: xarray.DataArray
        Input value.
    dim: list of strs (optional)
        Dimensions to apply the function over.
        If not given the function will be applied over all dimensions.

    Returns
    -------
    vmax: xarray.DataArray or float
        The result of the maximum value over the dimensions.

    """
    # NUMPY: replace by np.max(x, axis=axis) put directly in the file
    # float returned if the function is applied over all the dimensions
    if dim is None or set(x.dims) == set(dim):
        return float(x.max())

    return x.max(dim=dim)


def invert_matrix(mat):
    """
    Implements Vensim's INVERT MATRIX function.

    Invert the matrix defined by the last two dimensions of xarray.DataArray.

    Parameters
    -----------
    mat: xarray.DataArray
        The matrix to invert.

    Returns
    -------
    mat1: xarray.DataArray
        Inverted matrix.

    """
    # NUMPY: avoid converting to xarray, put directly the expression
    # in the model
    return xr.DataArray(np.linalg.inv(mat.values), mat.coords, mat.dims)


def vector_select(selection_array, expression_array, dim,
                  missing_vals, numerical_action, error_action):
    """
    Implements Vensim's VECTOR SELECT function.
    http://vensim.com/documentation/fn_vector_select.html

    Parameters
    ----------
    selection_array: xr.DataArray
        This specifies a selection array with a mixture of zeroes and
        non-zero values.
    expression_array: xarray.DataArray
        This is the expression that elements are being selected from
        based on the selection array.
    dim: list of strs
        Dimensions to apply the function over.
    missing_vals: float
        The value to use in the case where there are only zeroes in the
        selection array.
    numerical_action: int
        The action to take:
            - 0 It will calculate the weighted sum.
            - 1 When values in the selection array are non-zero, this
              will calculate the product of the
              selection_array * expression_array.
            - 2 The weighted minimum, for non zero values of the
              selection array, this is minimum of
              selection_array * expression_array.
            - 3 The weighted maximum, for non zero values of the
              selection array, this is maximum of
              selection_array * expression_array.
            - 4 For non zero values of the selection array, this is
              the average of selection_array * expression_array.
            - 5 When values in the selection array are non-zero,
              this will calculate the product of the
              expression_array ^ selection_array.
            - 6 When values in the selection array are non-zero,
              this will calculate the sum of the expression_array.
              The same as the SUM function for non-zero values in
              the selection array.
            - 7 When values in the selection array are non-zero,
              this will calculate the product of the expression_array.
              The same as the PROD function for non-zero values in
              the selection array.
            - 8 The unweighted minimum, for non zero values of the
              selection array, this is minimum of the expression_array.
              The same as the VMIN function for non-zero values in
              the selection array.
            - 9 The unweighted maximum, for non zero values of the
              selection array, this is maximum of expression_array.
              The same as the VMAX function for non-zero values in
              the selection array.
            - 10 For non zero values of the selection array,
              this is the average of expression_array.
    error_action: int
        Indicates how to treat too many or too few entries in the selection:
            - 0 No error is raised.
            - 1 Raise a floating point error is selection array only
              contains zeros.
            - 2 Raise an error if the selection array contains more
              than one non-zero value.
            - 3 Raise an error if all elements in selection array are
              zero, or more than one element is non-zero
              (this is a combination of error_action = 1 and error_action = 2).

    Returns
    -------
    result: xarray.DataArray or float
        The output of the numerical action.

    """
    zeros = (selection_array == 0).all(dim=dim)
    non_zeros = (selection_array != 0).sum(dim=dim)

    # Manage error actions
    if np.any(zeros) and error_action in (1, 3):
        raise FloatingPointError(
            "All the values of selection_array are 0...")

    if np.any(non_zeros > 1) and error_action in (2, 3):
        raise FloatingPointError(
            "More than one non-zero values in selection_array...")

    # Manage numeric actions (array to operate)
    # NUMPY: replace by np.where
    if numerical_action in range(5):
        array = xr.where(
            selection_array == 0,
            np.nan,
            selection_array * expression_array
        )
    elif numerical_action == 5:
        warnings.warn(
            "Vensim's help says that numerical_action=5 computes the "
            "product of selection_array ^ expression_array. But, in fact,"
            " Vensim is computing the product of expression_array ^ "
            " selection array. The output of this function behaves as "
            "Vensim, expression_array ^ selection_array."
        )
        array = xr.where(
            selection_array == 0,
            np.nan,
            expression_array ** selection_array
        )
    elif numerical_action in range(6, 11):
        array = xr.where(
            selection_array == 0,
            np.nan,
            expression_array
        )
    else:
        raise ValueError(
            f"Invalid argument value 'numerical_action={numerical_action}'. "
            "'numerical_action' must be 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 or 10.")

    # Manage numeric actions (operation)
    # NUMPY: use the axis
    if numerical_action in (0, 6):
        out = array.sum(dim=dim, skipna=True)
    elif numerical_action in (1, 5, 7):
        out = array.prod(dim=dim, skipna=True)
    elif numerical_action in (2, 8):
        out = array.min(dim=dim, skipna=True)
    elif numerical_action in (3, 9):
        out = array.max(dim=dim, skipna=True)
    elif numerical_action in (4, 10):
        out = array.mean(dim=dim, skipna=True)

    # Replace missin vals
    if len(out.shape) == 0 and np.all(zeros):
        return missing_vals
    elif len(out.shape) == 0:
        return float(out)
    elif np.any(zeros):
        out.values[zeros.values] = missing_vals

    return out


def vector_sort_order(vector, direction):
    """
    Implements Vensim's VECTOR SORT ORDER function. Sorting is done on
    the complete vector relative to the last subscript.
    https://www.vensim.com/documentation/fn_vector_sort_order.html

    Parameters
    -----------
    vector: xarray.DataArray
        The vector to sort.
    direction: float
        The direction to sort the vector. If direction > 0 it will sort
        the vector entries from smallest to biggest, otherwise from
        biggest to smallest.

    Returns
    -------
    vector_sorted: xarray.DataArray
        The sorted vector.

    """
    # TODO: can direction be an array? In this case this will fail
    if direction <= 0:
        # NUMPY: return flip directly
        flip = np.flip(vector.argsort(), axis=-1)
        return xr.DataArray(flip.values, vector.coords, vector.dims)
    return vector.argsort()


def vector_reorder(vector, svector):
    """
    Implements Vensim's VECTOR REORDER function. Reordering is done on
    the complete vector relative to the last subscript.
    https://www.vensim.com/documentation/fn_vector_reorder.html

    Parameters
    -----------
    vector: xarray.DataArray
        The vector to sort.
    svector: xarray.DataArray
        The vector to specify the order.

    Returns
    -------
    vector_sorted: xarray.DataArray
        The sorted vector.

    """
    # NUMPY: Use directly numpy sort functions, no need to assign coords later
    if len(svector.dims) > 1:
        # TODO this may be simplified
        new_vector = vector.copy()
        dims = svector.dims
        # create an empty array to hold the orderings (only last dim)
        arr = xr.DataArray(
            np.nan,
            {dims[-1]: vector.coords[dims[-1]].values},
            dims[-1:]
        )
        # split the ordering array in 0-dim arrays
        svectors = utils.xrsplit(svector)
        orders = {}
        for sv in svectors:
            # regrup the ordering arrays using last dimensions
            pos = {dim: str(sv.coords[dim].values) for dim in dims[:-1]}
            key = ";".join(pos.values())
            if key not in orders.keys():
                orders[key] = (pos, arr.copy())
            orders[key][1].loc[sv.coords[dims[-1]]] = sv.values

        for pos, array in orders.values():
            # get the reordered array
            values = [vector.loc[pos].values[int(i)] for i in array.values]
            new_vector.loc[pos] = values

        return new_vector

    return vector[svector.values].assign_coords(vector.coords)


def vector_rank(vector, direction):
    """
    Implements Vensim's VECTOR RANK function. Ranking is done on the
    complete vector relative to the last subscript.
    https://www.vensim.com/documentation/fn_vector_rank.html

    Parameters
    -----------
    vector: xarray.DataArray
        The vector to sort.
    direction: float
        The direction to sort the vector. If direction > 1 it will rank
        the vector entries from smallest to biggest, otherwise from
        biggest to smallest.

    Returns
    -------
    vector_rank: xarray.DataArray
        The rank of the vector.

    """
    return vector_sort_order(vector, direction).argsort() + 1


def get_time_value(time, relativeto, offset, measure):
    """
    Implements Vensim's GET TIME VALUE function. Warning, not all the
    cases are implemented.
    https://www.vensim.com/documentation/fn_get_time_value.html

    Parameters
    ----------
    time: pysd.py_backend.components.Time
        Model time object.
    relativeto: int
        The time to take as a reference:
            - 0 for the current simulation time.
            - 1 for the initial simulation time.
            - 2 for the current computer clock time.
    offset: float or xarray.DataArray
        The difference in time, as measured in the units of Time for
        the  model, to move before computing the value. offset is
        ignored when relativeto is 2.
    measure: int
        The units or measure of time:
            - 0 units of Time in the model (only for relativeto 0 and 1)
            - 1 years since 1 BC (an integer, same as the normal calendar year)
            - 2 quarter of year (1-4)
            - 3 month of year (1-12)
            - 4 day of month (1-31)
            - 5 day of week (0-6 where 0 is Sunday)
            - 6 days since Jan 1., 1 BC (year 1 BC is treated as year 0)
            - 7 hour of day (0-23)
            - 8 minute of hour (0-59)
            - 9 second of minute (0-59.999999 – not an integer)
            - 10 elapsed seconds modulo 500,000 (0-499,999)

    Returns
    -------
    time_value: float or int
        The resulting time value.

    """
    if relativeto == 0:
        # Current time
        ctime = time()
    elif relativeto == 1:
        # Initial time
        # Not implemented as it doesn't work as Vensim docs say
        # TODO check other versions or implement it as it should be?
        raise NotImplementedError("'relativeto=1' not implemented...")
        # ctime = time.initial_time()
    elif relativeto == 2:
        # Machine time
        ctime = utils.get_current_computer_time()
    else:
        # Invalid value
        raise ValueError(
            f"Invalid argument value 'relativeto={relativeto}'. "
            "'relativeto' must be 0, 1 or 2.")

    if measure == 0:
        # units of Time in the model (only for relativeto 0 and 1)
        if relativeto == 2:
            # measure=0 only supported with relativeto=0 or 1
            raise ValueError(
                "Invalid argument 'measure=0' with 'relativeto=2'.")
        else:
            return ctime + offset
    elif measure == 1:
        # years since 1 BC (an integer, same as the normal calendar year)
        if relativeto == 2:
            return ctime.year
    elif measure == 2:
        # quarter of year (1-4)
        if relativeto == 2:
            return int(1 + (ctime.month-0.5) // 3)
    elif measure == 3:
        # month of year (1-12)
        if relativeto == 2:
            return ctime.month
    elif measure == 4:
        # day of month (1-31)
        if relativeto == 2:
            return ctime.day
    elif measure == 5:
        # day of week (0-6 where 0 is Sunday)
        if relativeto == 2:
            return ctime.weekday()
    elif measure == 6:
        # days since Jan 1., 1 BC (year 1 BC is treated as year 0)
        if relativeto == 2:
            return (ctime - datetime(1, 1, 1)).days
    elif measure == 7:
        # hour of day (0-23)
        if relativeto == 2:
            return ctime.hour
    elif measure == 8:
        # minute of hour (0-59)
        if relativeto == 2:
            return ctime.minute
    elif measure == 9:
        # second of minute (0-59.99 – not an integer)
        if relativeto == 2:
            return ctime.second + 1e-6*ctime.microsecond
    elif measure == 10:
        # elapsed seconds modulo 500,000 (0-499,999)
        if relativeto == 2:
            return (ctime - datetime(1, 1, 1)).seconds % 500000
    else:
        raise ValueError(
            f"Invalid argument value 'measure={measure}'. "
            "'measure' must be 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 or 10.")

    # TODO include other measures for relativeto=0
    raise NotImplementedError(
        f"The case 'relativeto={relativeto}' and 'measure={measure}' "
        "is not implemented..."
    )



"""
The Stateful objects are used and updated each time step with an update
method. This include Integs, Delays, Forecasts, Smooths, and Trends,
between others. The Macro class and Model class are also Stateful type.
However, they are defined appart as they are more complex.
"""
import warnings

import numpy as np
import xarray as xr

from .functions import zidz, if_then_else


SMALL_VENSIM = 1e-6  # What is considered zero according to Vensim Help


class Stateful(object):
    # the integrator needs to be able to 'get' the current state of the object,
    # and get the derivative. It calculates the new state, and updates it.
    # The state can be any object which is subject to basic (element-wise)
    # algebraic operations
    def __init__(self):
        self._state = None
        self.shape_info = None
        self.py_name = ""

    def __call__(self, *args, **kwargs):
        return self.state

    @property
    def state(self):
        if self._state is None:
            raise AttributeError(
                self.py_name
                + "\nAttempt to call stateful element"
                + " before it is initialized.")
        return self._state

    @state.setter
    def state(self, new_value):
        if self.shape_info:
            self._state = xr.DataArray(data=new_value, **self.shape_info)
        else:
            self._state = new_value


class DynamicStateful(Stateful):

    def __init__(self):
        super().__init__()

    def update(self, state):
        try:
            self.state = state
        except Exception as err:
            raise ValueError(err.args[0] + "\n\n"
                             + "Could not update the value of "
                             + self.py_name)


class Integ(DynamicStateful):
    """
    Implements INTEG function.

    Parameters
    ----------
    ddt: callable
        Derivate to integrate.
    initial_value: callable
        Initial value.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: float or xarray.DataArray
        Current state of the object. Value of the stock.

    """
    def __init__(self, ddt, initial_value, py_name):
        super().__init__()
        self.init_func = initial_value
        self.ddt = ddt
        self.shape_info = None
        self.py_name = py_name

    def initialize(self, init_val=None):
        if init_val is None:
            self.state = self.init_func()
        else:
            self.state = init_val
        if isinstance(self.state, xr.DataArray):
            self.shape_info = {'dims': self.state.dims,
                               'coords': self.state.coords}

    def export(self):
        return {'state': self.state, 'shape_info': self.shape_info}


class NonNegativeInteg(Integ):
    """
    Implements non negative INTEG function.

    Parameters
    ----------
    ddt: callable
        Derivate to integrate.
    initial_value: callable
        Initial value.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: float or xarray.DataArray
        Current state of the object. Value of the stock.

    """
    def __init__(self, ddt, initial_value, py_name):
        super().__init__(ddt, initial_value, py_name)

    def update(self, state):
        self.state = np.maximum(state, 0)


class Delay(DynamicStateful):
    """
    Implements DELAY function.

    Parameters
    ----------
    delay_input: callable
        Input of the delay.
    delay_time: callable
        Delay time.
    initial_value: callable
        Initial value.
    order: callable
        Delay order.
    tsetp: callable
        The time step of the model.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: numpy.array or xarray.DataArray
        Current state of the object. Array of the delays values multiplied
        by their corresponding average time.

    """
    # note that we could have put the `delay_input` argument as a parameter to
    # the `__call__` function, and more closely mirrored the vensim syntax.
    # However, people may get confused this way in thinking that they need
    # only one delay object and can call it with various arguments to delay
    # whatever is convenient. This method forces them to acknowledge that
    # additional structure is being created in the delay object.

    def __init__(self, delay_input, delay_time, initial_value, order, tstep,
                 py_name):
        super().__init__()
        self.init_func = initial_value
        self.delay_time_func = delay_time
        self.input_func = delay_input
        self.order_func = order
        self.order = None
        self.tstep = tstep
        self.shape_info = None
        self.py_name = py_name

    def initialize(self, init_val=None):
        order = self.order_func()

        if order != int(order):
            warnings.warn(self.py_name + '\n' +
                          'Casting delay order '
                          + f'from {order} to {int(order)}')

        self.order = int(order)  # The order can only be set once
        if self.order*self.tstep() > np.min(self.delay_time_func()):
            while self.order*self.tstep() > np.min(self.delay_time_func()):
                self.order -= 1
            warnings.warn(self.py_name + '\n' +
                          'Delay time very small, casting delay order '
                          + f'from {int(order)} to {self.order}')

        if init_val is None:
            init_state_value = self.init_func() * self.delay_time_func()
        else:
            init_state_value = init_val * self.delay_time_func()

        if isinstance(init_state_value, xr.DataArray):
            # broadcast self.state
            self.state = init_state_value.expand_dims({
                '_delay': np.arange(self.order)}, axis=0)
            self.shape_info = {'dims': self.state.dims,
                               'coords': self.state.coords}
        else:
            self.state = np.array([init_state_value] * self.order)

    def __call__(self):
        if self.shape_info:
            return self.state[-1].reset_coords('_delay', drop=True)\
                   / self.delay_time_func()
        else:
            return self.state[-1] / self.delay_time_func()

    def ddt(self):
        outflows = self.state / self.delay_time_func()
        inflows = np.roll(outflows, 1, axis=0)
        if self.shape_info:
            inflows[0] = self.input_func().values
        else:
            inflows[0] = self.input_func()
        return (inflows - outflows) * self.order

    def export(self):
        return {'state': self.state, 'shape_info': self.shape_info}


class DelayN(DynamicStateful):
    """
    Implements DELAY N function.

    Parameters
    ----------
    delay_input: callable
        Input of the delay.
    delay_time: callable
        Delay time.
    initial_value: callable
        Initial value.
    order: callable
        Delay order.
    tsetp: callable
        The time step of the model.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: numpy.array or xarray.DataArray
        Current state of the object. Array of the delays values multiplied
        by their corresponding average time.

    times: numpy.array or xarray.DataArray
        Array of delay times used for computing the delay output.
        If delay_time is constant, this array will be constant and
        DelayN will behave ad Delay.

    """
    # note that we could have put the `delay_input` argument as a parameter to
    # the `__call__` function, and more closely mirrored the vensim syntax.
    # However, people may get confused this way in thinking that they need
    # only one delay object and can call it with various arguments to delay
    # whatever is convenient. This method forces them to acknowledge that
    # additional structure is being created in the delay object.

    def __init__(self, delay_input, delay_time, initial_value, order, tstep,
                 py_name):
        super().__init__()
        self.init_func = initial_value
        self.delay_time_func = delay_time
        self.input_func = delay_input
        self.order_func = order
        self.order = None
        self.times = None
        self.tstep = tstep
        self.shape_info = None
        self.py_name = py_name

    def initialize(self, init_val=None):
        order = self.order_func()

        if order != int(order):
            warnings.warn(self.py_name + '\n' +
                          'Casting delay order '
                          + f'from {order} to {int(order)}')

        self.order = int(order)  # The order can only be set once
        if self.order*self.tstep() > np.min(self.delay_time_func()):
            while self.order*self.tstep() > np.min(self.delay_time_func()):
                self.order -= 1
            warnings.warn(self.py_name + '\n' +
                          'Delay time very small, casting delay order '
                          + f'from {int(order)} to {self.order}')

        if init_val is None:
            init_state_value = self.init_func() * self.delay_time_func()
        else:
            init_state_value = init_val * self.delay_time_func()

        if isinstance(init_state_value, xr.DataArray):
            # broadcast self.state
            self.state = init_state_value.expand_dims({
                '_delay': np.arange(self.order)}, axis=0)
            self.times = self.delay_time_func().expand_dims({
                '_delay': np.arange(self.order)}, axis=0)
            self.shape_info = {'dims': self.state.dims,
                               'coords': self.state.coords}
        else:
            self.state = np.array([init_state_value] * self.order)
            self.times = np.array([self.delay_time_func()] * self.order)

    def __call__(self):
        if self.shape_info:
            return self.state[-1].reset_coords('_delay', drop=True)\
                   / self.times[0].reset_coords('_delay', drop=True)
        else:
            return self.state[-1] / self.times[0]

    def ddt(self):
        if self.shape_info:
            # if is xarray need to preserve coords
            self.times = self.times.roll({'_delay': 1}, False)
            self.times[0] = self.delay_time_func()
            outflows = self.state / self.times
            inflows = outflows.roll({'_delay': 1}, False)
        else:
            # if is float use numpy.roll
            self.times = np.roll(self.times, 1, axis=0)
            self.times[0] = self.delay_time_func()
            outflows = self.state / self.times
            inflows = np.roll(outflows, 1, axis=0)

        inflows[0] = self.input_func()
        return (inflows - outflows)*self.order

    def export(self):
        return {'state': self.state, 'times': self.times,
                'shape_info': self.shape_info}


class DelayFixed(DynamicStateful):
    """
    Implements DELAY FIXED function.

    Parameters
    ----------
    delay_input: callable
        Input of the delay.
    delay_time: callable
        Delay time.
    initial_value: callable
        Initial value.
    tsetp: callable
        The time step of the model.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: float or xarray.DataArray
        Current state of the object, equal to pipe[pointer].
    pipe: list
        List of the delays values.
    pointer: int
        Pointer to the last value in the pipe

    """

    def __init__(self, delay_input, delay_time, initial_value, tstep,
                 py_name):
        super().__init__()
        self.init_func = initial_value
        self.delay_time_func = delay_time
        self.input_func = delay_input
        self.tstep = tstep
        self.order = None
        self.pointer = 0
        self.py_name = py_name

    def initialize(self, init_val=None):
        order = max(self.delay_time_func()/self.tstep(), 1)

        if order != int(order):
            warnings.warn(
                self.py_name + '\n'
                + 'Casting delay order from %f to %i' % (
                    order, round(order + SMALL_VENSIM)))

        # need to add a small decimal to ensure that 0.5 is rounded to 1
        # The order can only be set once
        self.order = round(order + SMALL_VENSIM)

        # set the pointer to 0
        self.pointer = 0

        if init_val is None:
            init_state_value = self.init_func()
        else:
            init_state_value = init_val

        self.state = init_state_value
        self.pipe = [init_state_value] * self.order

    def __call__(self):
        return self.state

    def ddt(self):
        return np.nan

    def update(self, state):
        self.pipe[self.pointer] = self.input_func()
        self.pointer = (self.pointer + 1) % self.order
        self.state = self.pipe[self.pointer]

    def export(self):
        return {'state': self.state, 'pointer': self.pointer,
                'pipe': self.pipe}


class Forecast(DynamicStateful):
    """
    Implements FORECAST function.

    Parameters
    ----------
    forecast_input: callable
        Input of the forecast.
    average_time: callable
        Average time.
    horizon: callable
        Forecast horizon.
    initial_trend: callable
        Initial trend of the forecast.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: float or xarray.DataArray
        Current state of the object. AV value by Vensim docs.

    """
    def __init__(self, forecast_input, average_time, horizon, initial_trend,
                 py_name):
        super().__init__()
        self.horizon = horizon
        self.average_time = average_time
        self.input = forecast_input
        self.initial_trend = initial_trend
        self.py_name = py_name

    def initialize(self, init_trend=None):

        # self.state = AV in the vensim docs
        if init_trend is None:
            self.state = self.input() / (1 + self.initial_trend())
        else:
            self.state = self.input() / (1 + init_trend)

        if isinstance(self.state, xr.DataArray):
            self.shape_info = {'dims': self.state.dims,
                               'coords': self.state.coords}

    def __call__(self):
        return self.input() * (
            1 + zidz(self.input() - self.state,
                     self.average_time() * self.state
                     )*self.horizon()
        )

    def ddt(self):
        return (self.input() - self.state) / self.average_time()

    def export(self):
        return {'state': self.state, 'shape_info': self.shape_info}


class Smooth(DynamicStateful):
    """
    Implements SMOOTH function.

    Parameters
    ----------
    smooth_input: callable
        Input of the smooth.
    smooth_time: callable
        Smooth time.
    initial_value: callable
        Initial value.
    order: callable
        Delay order.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: numpy.array or xarray.DataArray
        Current state of the object. Array of the inputs having the
        value to return in the last position.

    """
    def __init__(self, smooth_input, smooth_time, initial_value, order,
                 py_name):
        super().__init__()
        self.init_func = initial_value
        self.smooth_time_func = smooth_time
        self.input_func = smooth_input
        self.order_func = order
        self.order = None
        self.shape_info = None
        self.py_name = py_name

    def initialize(self, init_val=None):
        self.order = self.order_func()  # The order can only be set once

        if init_val is None:
            init_state_value = self.init_func()
        else:
            init_state_value = init_val

        if isinstance(init_state_value, xr.DataArray):
            # broadcast self.state
            self.state = init_state_value.expand_dims({
                '_smooth': np.arange(self.order)}, axis=0)
            self.shape_info = {'dims': self.state.dims,
                               'coords': self.state.coords}
        else:
            self.state = np.array([init_state_value] * self.order)

    def __call__(self):
        if self.shape_info:
            return self.state[-1].reset_coords('_smooth', drop=True)
        else:
            return self.state[-1]

    def ddt(self):
        targets = np.roll(self.state, 1, axis=0)
        if self.shape_info:
            targets[0] = self.input_func().values
        else:
            targets[0] = self.input_func()
        return (targets - self.state) * self.order / self.smooth_time_func()

    def export(self):
        return {'state': self.state, 'shape_info': self.shape_info}


class Trend(DynamicStateful):
    """
    Implements TREND function.

    Parameters
    ----------
    trend_input: callable
        Input of the trend.
    average_time: callable
        Average time.
    initial_trend: callable
        Initial trend.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: float or xarray.DataArray
        Current state of the object. AV value by Vensim docs.

    """
    def __init__(self, trend_input, average_time, initial_trend, py_name):
        super().__init__()
        self.init_func = initial_trend
        self.average_time_function = average_time
        self.input_func = trend_input
        self.py_name = py_name

    def initialize(self, init_trend=None):
        if init_trend is None:
            self.state = self.input_func()\
                / (1 + self.init_func()*self.average_time_function())
        else:
            self.state = self.input_func()\
                / (1 + init_trend*self.average_time_function())

        if isinstance(self.state, xr.DataArray):
            self.shape_info = {'dims': self.state.dims,
                               'coords': self.state.coords}

    def __call__(self):
        return zidz(self.input_func() - self.state,
                    self.average_time_function() * np.abs(self.state))

    def ddt(self):
        return (self.input_func() - self.state) / self.average_time_function()

    def export(self):
        return {'state': self.state, 'shape_info': self.shape_info}


class SampleIfTrue(DynamicStateful):
    """
    Implements SAMPLE IF TRUE function.

    Parameters
    ----------
    condition: callable
        Condition for sample.
    actual_value: callable
        Value to update if condition is true.
    initial_value: callable
        Initial value.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: float or xarray.DataArray
        Current state of the object. Last actual_value when condition
        was true or the initial_value if condition has never been true.

    """
    def __init__(self, condition, actual_value, initial_value, py_name):
        super().__init__()
        self.condition = condition
        self.actual_value = actual_value
        self.init_func = initial_value
        self.py_name = py_name

    def initialize(self, init_val=None):
        if init_val is None:
            self.state = self.init_func()
        else:
            self.state = init_val
        if isinstance(self.state, xr.DataArray):
            self.shape_info = {'dims': self.state.dims,
                               'coords': self.state.coords}

    def __call__(self):
        return if_then_else(self.condition(),
                            self.actual_value,
                            lambda: self.state)

    def ddt(self):
        return np.nan

    def update(self, state):
        self.state = self.state*0 + if_then_else(self.condition(),
                                                 self.actual_value,
                                                 lambda: self.state)

    def export(self):
        return {'state': self.state, 'shape_info': self.shape_info}


class Initial(Stateful):
    """
    Implements INITIAL function.

    Parameters
    ----------
    initial_value: callable
        Initial value.
    py_name: str
        Python name to identify the object.

    Attributes
    ----------
    state: float or xarray.DataArray
        Current state of the object, which will always be the initial_value.

    """
    def __init__(self, initial_value, py_name):
        super().__init__()
        self.init_func = initial_value
        self.py_name = py_name

    def initialize(self, init_val=None):
        if init_val is None:
            self.state = self.init_func()
        else:
            self.state = init_val

    def export(self):
        return {'state': self.state}



"""
pysd.py

Contains all the code that will be directly accessed by the user in
normal operation.
"""

import sys
from warnings import warn

from pysd.py_backend.model import Model


if sys.version_info[:2] < (3, 9):  # pragma: no cover
    raise RuntimeError(
        "\n\n"
        + "Your Python version is no longer supported by PySD.\n"
        + "The current version needs to run at least Python 3.9."
        + " You are running:\n\tPython "
        + sys.version
        + "\nPlease update your Python version or use the last "
        + " supported version:\n\t"
        + "https://github.com/SDXorg/pysd/releases/tag/LastPy2"
    )


def read_xmile(xmile_file, data_files=None, initialize=True,
               missing_values="warning"):
    """
    Construct a model from a Xmile file.

    Parameters
    ----------
    xmile_file:  str or pathlib.Path
        The relative path filename for a raw Xmile file.

    initialize: bool (optional)
        If False, the model will not be initialize when it is loaded.
        Default is True.

    data_files: list or str or None (optional)
        If given the list of files where the necessary data to run the model
        is given. Default is None.

    missing_values: str ("warning", "error", "ignore", "keep") (optional)
        What to do with missing values. If "warning" (default)
        shows a warning message and interpolates the values.
        If "raise" raises an error. If "ignore" interpolates
        the values without showing anything. If "keep" it will keep
        the missing values, this option may cause the integration to
        fail, but it may be used to check the quality of the data.

    Returns
    -------
    model: a PySD class object
        Elements from the Python model are loaded into the PySD class
        and ready to run

    Examples
    --------
    >>> model = read_xmile('../tests/test-models/samples/teacup/teacup.xmile')

    """
    from pysd.translators.xmile.xmile_file import XmileFile
    from pysd.builders.python.python_model_builder import ModelBuilder

    # Read and parse Xmile file
    xmile_file_obj = XmileFile(xmile_file)
    xmile_file_obj.parse()

    # get AbstractModel
    abs_model = xmile_file_obj.get_abstract_model()

    # build Python file
    py_model_file = ModelBuilder(abs_model).build_model()

    # load Python file
    model = load(py_model_file, data_files, initialize, missing_values)
    model.xmile_file = str(xmile_file)

    return model


def read_vensim(mdl_file, data_files=None, initialize=True,
                missing_values="warning", split_views=False,
                encoding=None, **kwargs):
    """
    Construct a model from Vensim `.mdl` file.

    Parameters
    ----------
    mdl_file: str or pathlib.Path
        The relative path filename for a raw Vensim `.mdl` file.

    initialize: bool (optional)
        If False, the model will not be initialize when it is loaded.
        Default is True.

    data_files: list or str or None (optional)
        If given the list of files where the necessary data to run the model
        is given. Default is None.

    missing_values: str ("warning", "error", "ignore", "keep") (optional)
        What to do with missing values. If "warning" (default)
        shows a warning message and interpolates the values.
        If "raise" raises an error. If "ignore" interpolates
        the values without showing anything. If "keep" it will keep
        the missing values, this option may cause the integration to
        fail, but it may be used to check the quality of the data.

    split_views: bool (optional)
        If True, the sketch is parsed to detect model elements in each
        model view, and then translate each view in a separate Python
        file. Setting this argument to True is recommended for large
        models split in many different views. Default is False.

    encoding: str or None (optional)
        Encoding of the source model file. If None, the encoding will be
        read from the model, if the encoding is not defined in the model
        file it will be set to 'UTF-8'. Default is None.

    subview_sep: list
        Characters used to separate views and subviews (e.g. [",", "."]).
        If provided, and split_views=True, each submodule will be placed
        inside the directory of the parent view.

    **kwargs: (optional)
        Additional keyword arguments for translation.

    Returns
    -------
    model: a PySD class object
        Elements from the Python model are loaded into the PySD class
        and ready to run

    Examples
    --------
    >>> model = read_vensim('../tests/test-models/samples/teacup/teacup.mdl')

    """
    from pysd.translators.vensim.vensim_file import VensimFile
    from pysd.builders.python.python_model_builder import ModelBuilder
    # Read and parse Vensim file
    ven_file = VensimFile(mdl_file, encoding=encoding)
    ven_file.parse()
    if split_views:
        # split variables per views
        subview_sep = kwargs.get("subview_sep", "")
        ven_file.parse_sketch(subview_sep)

    # get AbstractModel
    abs_model = ven_file.get_abstract_model()

    # build Python file
    py_model_file = ModelBuilder(abs_model).build_model()

    # load Python file
    model = load(py_model_file, data_files, initialize, missing_values)
    model.mdl_file = str(mdl_file)

    return model


def load(py_model_file, data_files=None, initialize=True,
         missing_values="warning"):
    """
    Load a Python-converted model file.

    Parameters
    ----------
    py_model_file : str
        Filename of a model which has already been converted into a
        Python format.

    initialize: bool (optional)
        If False, the model will not be initialize when it is loaded.
        Default is True.

    data_files: dict or list or str or None
        The dictionary with keys the name of file and variables to
        load the data from there. Or the list of names or name of the
        file to search the data in. Only works for TabData type object
        and it is neccessary to provide it. Default is None.

    missing_values : str ("warning", "error", "ignore", "keep") (optional)
        What to do with missing values. If "warning" (default)
        shows a warning message and interpolates the values.
        If "raise" raises an error. If "ignore" interpolates
        the values without showing anything. If "keep" it will keep
        the missing values, this option may cause the integration to
        fail, but it may be used to check the quality of the data.

    Examples
    --------
    >>> model = load('../tests/test-models/samples/teacup/teacup.py')

    """
    return Model(py_model_file, data_files, initialize, missing_values)




# functions that can be diretcly applied over an array
functionspace = {
    # directly build functions without dependencies
    "elmcount": ("len(%(0)s)", ()),

    # directly build numpy based functions
    "pi": ("np.pi", (("numpy",),)),
    "abs": ("np.abs(%(0)s)", (("numpy",),)),
    "power": ("np.power(%(0)s,%(1)s)", (("numpy",),)),
    "min": ("np.minimum(%(0)s, %(1)s)", (("numpy",),)),
    "max": ("np.maximum(%(0)s, %(1)s)", (("numpy",),)),
    "exp": ("np.exp(%(0)s)", (("numpy",),)),
    "sin": ("np.sin(%(0)s)", (("numpy",),)),
    "cos": ("np.cos(%(0)s)", (("numpy",),)),
    "tan": ("np.tan(%(0)s)", (("numpy",),)),
    "arcsin": ("np.arcsin(%(0)s)", (("numpy",),)),
    "arccos": ("np.arccos(%(0)s)", (("numpy",),)),
    "arctan": ("np.arctan(%(0)s)", (("numpy",),)),
    "sinh": ("np.sinh(%(0)s)", (("numpy",),)),
    "cosh": ("np.cosh(%(0)s)", (("numpy",),)),
    "tanh": ("np.tanh(%(0)s)", (("numpy",),)),
    "sqrt": ("np.sqrt(%(0)s)", (("numpy",),)),
    "ln": ("np.log(%(0)s)", (("numpy",),)),
    "log": ("(np.log(%(0)s)/np.log(%(1)s))", (("numpy",),)),
    # NUMPY: "invert_matrix": ("np.linalg.inv(%(0)s)", (("numpy",),)),

    # vector functions with axis to apply over
    # NUMPY:
    # "prod": "np.prod(%(0)s, axis=%(axis)s)", (("numpy",),)),
    # "sum": "np.sum(%(0)s, axis=%(axis)s)", (("numpy",),)),
    # "vmax": "np.max(%(0)s, axis=%(axis)s)", ("numpy",),)),
    # "vmin": "np.min(%(0)s, axis=%(axis)s)", (("numpy",),))
    "prod": ("prod(%(0)s, dim=%(axis)s)", (("functions", "prod"),)),
    "sum": ("sum(%(0)s, dim=%(axis)s)", (("functions", "sum"),)),
    "vmax": ("vmax(%(0)s, dim=%(axis)s)", (("functions", "vmax"),)),
    "vmin": ("vmin(%(0)s, dim=%(axis)s)", (("functions", "vmin"),)),
    "vmax_xmile": ("vmax(%(0)s)", (("functions", "vmax"),)),
    "vmin_xmile": ("vmin(%(0)s)", (("functions", "vmin"),)),
    "vector_select": (
        "vector_select(%(0)s, %(1)s, %(axis)s, %(2)s, %(3)s, %(4)s)",
        (("functions", "vector_select"),)
    ),

    # functions defined in pysd.py_bakcend.functions
    "active_initial": (
        "active_initial(__data[\"time\"].stage, lambda: %(0)s, %(1)s)",
        (("functions", "active_initial"),)),
    "if_then_else": (
        "if_then_else(%(0)s, lambda: %(1)s, lambda: %(2)s)",
        (("functions", "if_then_else"),)),
    "integer": (
        "integer(%(0)s)",
        (("functions", "integer"),)),
    "invert_matrix": (  # NUMPY: remove
        "invert_matrix(%(0)s)",
        (("functions", "invert_matrix"),)),  # NUMPY: remove
    "modulo": (
        "modulo(%(0)s, %(1)s)",
        (("functions", "modulo"),)),
    "pulse": (
        "pulse(__data['time'], %(0)s, width=%(1)s)",
        (("functions", "pulse"),)),
    "Xpulse": (
        "pulse(__data['time'], %(0)s, magnitude=%(1)s)",
        (("functions", "pulse"),)),
    "pulse_train": (
        "pulse(__data['time'], %(0)s, repeat_time=%(1)s, width=%(2)s, "\
        "end=%(3)s)",
        (("functions", "pulse"),)),
    "Xpulse_train": (
        "pulse(__data['time'], %(0)s, repeat_time=%(1)s, magnitude=%(2)s)",
        (("functions", "pulse"),)),
    "get_time_value": (
        "get_time_value(__data['time'], %(0)s, %(1)s, %(2)s)",
        (("functions", "get_time_value"),)),
    "quantum": (
        "quantum(%(0)s, %(1)s)",
        (("functions", "quantum"),)),
    "Xramp": (
        "ramp(__data['time'], %(0)s, %(1)s)",
        (("functions", "ramp"),)),
    "ramp": (
        "ramp(__data['time'], %(0)s, %(1)s, %(2)s)",
        (("functions", "ramp"),)),
    "step": (
        "step(__data['time'], %(0)s, %(1)s)",
        (("functions", "step"),)),
    "xidz": (
        "xidz(%(0)s, %(1)s, %(2)s)",
        (("functions", "xidz"),)),
    "zidz": (
        "zidz(%(0)s, %(1)s)",
        (("functions", "zidz"),)),
    "vector_sort_order": (
        "vector_sort_order(%(0)s, %(1)s)",
        (("functions", "vector_sort_order"),)),
    "vector_reorder": (
        "vector_reorder(%(0)s, %(1)s)",
        (("functions", "vector_reorder"),)),
    "vector_rank": (
        "vector_rank(%(0)s, %(1)s)",
        (("functions", "vector_rank"),)),

    # random functions must have the shape of the component subscripts
    # most of them are shifted, scaled and truncated
    "random_0_1": (
        "np.random.uniform(0, 1, size=%(size)s)",
        (("numpy",),)),
    "random_uniform": (
        "np.random.uniform(%(0)s, %(1)s, size=%(size)s)",
        (("numpy",),)),
    "random_normal": (
        "stats.truncnorm.rvs((%(0)s-%(2)s)/%(3)s, (%(1)s-%(2)s)/%(3)s,"
        " loc=%(2)s, scale=%(3)s, size=%(size)s)",
        (("scipy", "stats"),)),
    "random_exponential": (
        "stats.truncexpon.rvs((%(1)s-np.maximum(%(0)s, %(2)s))/%(3)s,"
        " loc=np.maximum(%(0)s, %(2)s), scale=%(3)s, size=%(size)s)",
        (("scipy", "stats"), ("numpy",),)),
}



import re
from ..structures import abstract_expressions as ae


structures = {
    "reference":  ae.ReferenceStructure,
    "subscripts_ref": ae.SubscriptsReferenceStructure,
    "arithmetic": ae.ArithmeticStructure,
    "logic": ae.LogicStructure,
    "inline_lookup": ae.InlineLookupsStructure,
    "lookup": ae.LookupsStructure,
    "call": ae.CallStructure,
    "init": ae.InitialStructure,
    "stock": ae.IntegStructure,
    "delay": {
        2: lambda x, y: ae.DelayFixedStructure(x, y, x),
        3: lambda x, y, z: ae.DelayFixedStructure(x, y, z)
    },
    "delay1": {
        2: lambda x, y: ae.DelayStructure(x, y, x, 1),
        3: lambda x, y, z: ae.DelayStructure(x, y, z, 1)
    },
    "delay3": {
        2: lambda x, y: ae.DelayStructure(x, y, x, 3),
        3: lambda x, y, z: ae.DelayStructure(x, y, z, 3),
    },
    "delayn": {
        3: lambda x, y, n: ae.DelayNStructure(x, y, x, n),
        4: lambda x, y, n, z: ae.DelayNStructure(x, y, z, n),
    },
    "smth1": {
        2: lambda x, y: ae.SmoothStructure(x, y, x, 1),
        3: lambda x, y, z: ae.SmoothStructure(x, y, z, 1)
    },
    "smth3": {
        2: lambda x, y: ae.SmoothStructure(x, y, x, 3),
        3: lambda x, y, z: ae.SmoothStructure(x, y, z, 3)
    },
    "smthn": {
        3: lambda x, y, n: ae.SmoothNStructure(x, y, x, n),
        4: lambda x, y, n, z: ae.SmoothNStructure(x, y, z, n)
    },
    "trend": {
        2: lambda x, y: ae.TrendStructure(x, y, 0),
        3: ae.TrendStructure,
    },
    "forcst": {
        3: lambda x, y, z: ae.ForecastStructure(x, y, z, 0),
        4: ae.ForecastStructure
    },
    "safediv": {
        2: lambda x, y: ae.CallStructure(
            ae.ReferenceStructure("zidz"), (x, y)),
        3: lambda x, y, z: ae.CallStructure(
            ae.ReferenceStructure("xidz"), (x, y, z))
    },
    "if_then_else": lambda x, y, z: ae.CallStructure(
            ae.ReferenceStructure("if_then_else"), (x, y, z)),
    "ramp": {
        2: lambda x, y: ae.CallStructure(
            ae.ReferenceStructure("Xramp"), (x, y)),
        3: lambda x, y, z: ae.CallStructure(
            ae.ReferenceStructure("ramp"), (x, y, z))
    },
    "pulse": {
        2: lambda magnitude, start: ae.CallStructure(
            ae.ReferenceStructure("Xpulse"), (start, magnitude)),
        3: lambda magnitude, start, interval: ae.CallStructure(
            ae.ReferenceStructure("Xpulse_train"), (start, interval, magnitude)
            )
    },
    "max": {
        1: lambda x: ae.CallStructure(
            ae.ReferenceStructure("vmax_xmile"), (x,)),
        2: lambda x, y: ae.CallStructure(
            ae.ReferenceStructure("max"), (x, y))
    },
    "min": {
        1: lambda x: ae.CallStructure(
            ae.ReferenceStructure("vmin_xmile"), (x,)),
        2: lambda x, y: ae.CallStructure(
            ae.ReferenceStructure("min"), (x, y))
    },
    "negative": lambda x: ae.ArithmeticStructure(["negative"], (x,)),
    "int": lambda x: ae.CallStructure(
            ae.ReferenceStructure("integer"), (x,))
}


operators = {
    "logic_ops": ["and", "or"],
    "not_ops": ["not"],
    "comp_ops": ["=", "<>", "<=", "<", ">=", ">"],
    "add_ops": ["+", "-"],
    "prod_ops": ["*", "/"],
    "exp_ops": ["^"],
    "pre_ops": ["+", "-"]
}


parsing_ops = {
    key: "|".join(re.escape(x) for x in values)
    for key, values in operators.items()
}



from typing import Union


class ImportsManager():
    """
    Class to save the imported modules information for intelligent import
    """
    _external_libs = {"numpy": "np", "xarray": "xr"}
    _external_submodules = ["scipy"]
    _internal_libs = [
        "functions", "statefuls", "external", "data", "lookups", "utils",
        "allocation", "model"
    ]

    def __init__(self):
        self._numpy, self._xarray = False, False
        self._functions, self._statefuls, self._external, self._data,\
            self._lookups, self._utils, self._scipy, self._allocation,\
            self._model =\
            set(), set(), set(), set(), set(), set(), set(), set(), set()

    def add(self, module: str, function: Union[str, None] = None) -> None:
        """
        Add a function from module.

        Parameters
        ----------
        module: str
          module name.

        function: str or None
          function name. If None module will be set to true.

        """
        if function:
            getattr(self, f"_{module}").add(function)
        else:
            setattr(self, f"_{module}", True)

    def get_header(self, outfile: str) -> str:
        """
        Returns the importing information to print in the model file

        Parameters
        ----------
        outfile: str
            Name of the outfile to print in the header.

        Returns
        -------
        text: str
            Header of the translated model file.

        """
        text =\
            f'"""\nPython model \'{outfile}\'\nTranslated using PySD\n"""\n\n'

        text += "from pathlib import Path\n"

        for module, shortname in self._external_libs.items():
            if getattr(self, f"_{module}"):
                text += f"import {module} as {shortname}\n"

        for module in self._external_submodules:
            if getattr(self, f"_{module}"):
                text += "from %(module)s import %(submodules)s\n" % {
                    "module": module,
                    "submodules": ", ".join(getattr(self, f"_{module}"))}

        text += "\n"

        for module in self._internal_libs:
            if getattr(self, f"_{module}"):
                text += "from pysd.py_backend.%(module)s import %(methods)s\n"\
                        % {
                            "module": module,
                            "methods": ", ".join(getattr(self, f"_{module}"))}

        text += "from pysd import Component\n"

        return text



"""
The XmileFile class allows reading the original Xmile model file and
parsing it into Section elements. The final result can be exported to an
AbstractModel class in order to build a model in another programming language.
"""
from typing import Union
from pathlib import Path
from lxml import etree

from ..structures.abstract_model import AbstractModel

from .xmile_section import Section
from .xmile_utils import supported_extensions


class XmileFile():
    """
    The XmileFile class allows parsing an Xmile file.
    When the object is created, the model file is automatically opened
    and parsed with lxml.etree.

    Parameters
    ----------
    xmile_path: str or pathlib.Path
        Path to the Xmile model.

    """
    def __init__(self, xmile_path: Union[str, Path]):
        self.xmile_path = Path(xmile_path)
        self.root_path = self.xmile_path.parent
        self.xmile_root = self._get_root()
        self.ns = self.xmile_root.nsmap[None]  # namespace of the xmile
        self.view_elements = None

    def __str__(self):  # pragma: no cover
        return "\nXmile model file, loaded from:\n\t%s\n" % self.xmile_path

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get model information."""
        text = self.__str__()
        for section in self.sections:
            text += section._verbose

        return text

    @property
    def verbose(self):  # pragma: no cover
        """Print model information to standard output."""
        print(self._verbose)

    def _get_root(self) -> etree._Element:
        """
        Read an Xmile file and assign its content to self.model_text

        Returns
        -------
        lxml.etree._Element: parsed xml object

        """
        # check for model extension
        if self.xmile_path.suffix.lower() not in supported_extensions:
            raise ValueError(
                "The file to translate, '%s' " % self.xmile_path
                + "is not a Xmile model. It must end with any of "
                + "%s extensions." % ', '.join(supported_extensions)
            )

        return etree.parse(
            str(self.xmile_path),
            parser=etree.XMLParser(encoding="utf-8", recover=True)
        ).getroot()

    def parse(self, parse_all: bool = True) -> None:
        """
        Create a XmileSection object from the model content and parse it.
        As macros are currently not supported, all models will
        have a single section. This function should split the macros in
        independent sections in the future.

        Parameters
        ----------
        parse_all: bool (optional)
            If True, the created XmileSection objects will be
            automatically parsed. Otherwise, these objects will only be
            added to self.sections but not parsed. Default is True.

        """
        # TODO: in order to make macros work we need to split them here
        # in several sections
        # We keep everything in a single section
        self.sections = [Section(
                name="__main__",
                path=self.xmile_path.with_suffix(".py"),
                section_type="main",
                params=[],
                returns=[],
                content_root=self.xmile_root,
                namespace=self.ns,
                split=False,
                views_dict=None)]

        if parse_all:
            for section in self.sections:
                section.parse()

    def get_abstract_model(self) -> AbstractModel:
        """
        Get Abstract Model used for building. This, method should be
        called after parsing the model (self.parse). This automatically
        calls the get_abstract_section method from the model sections.

        Returns
        -------
        AbstractModel: AbstractModel
          Abstract Model object that can be used for building the model
          in another language.

        """
        return AbstractModel(
            original_path=self.xmile_path,
            sections=tuple(section.get_abstract_section()
                           for section in self.sections))



import sys
import os
from pathlib import Path

from datetime import datetime

import pysd
from pysd.translators.vensim.vensim_utils import supported_extensions as\
    vensim_extensions
from pysd.translators.xmile.xmile_utils import supported_extensions as\
    xmile_extensions

from .parser import parser


def main(args):
    """
    Main function. Reads user arguments, loads the models,
    runs it and saves the output

    Parameters
    ----------
    args: list
        User arguments.

    Returns
    -------
    None

    """
    options = parser.parse_args(args)

    model = load(options.model_file, options.data_files,
                 options.missing_values, options.split_views,
                 subview_sep=options.subview_sep)

    if not options.run:
        print("\nFinished!")
        sys.exit()

    model.initialize()

    if not options.output_file:
        options.output_file = os.path.splitext(os.path.basename(
            options.model_file
            ))[0]\
                + datetime.now().strftime("_output_%Y_%m_%d-%H_%M_%S_%f.tab")

    model.run(**create_configuration(model, options))

    if options.export_file:
        model.export(options.export_file)

    print("\nFinished!")
    sys.exit()


def load(model_file, data_files, missing_values, split_views, **kwargs):
    """
    Translate and load model file.

    Paramters
    ---------
    model_file: str
        Vensim, Xmile or PySD model file.

    data_files: list
        If given the list of files where the necessary data to run the model
        is given.

    missing_values : str ("warning", "error", "ignore", "keep")
        What to do with missing values. If "warning" (default)
        shows a warning message and interpolates the values.
        If "raise" raises an error. If "ignore" interpolates
        the values without showing anything. If "keep" it will keep
        the missing values, this option may cause the integration to
        fail, but it may be used to check the quality of the data.

    split_views: bool (optional)
        If True, the sketch is parsed to detect model elements in each
        model view, and then translate each view in a separate Python
        file. Setting this argument to True is recommended for large
        models split in many different views. Default is False.

    **kwargs: (optional)
        Additional keyword arguments.
        subview_sep:(str)
            Character used to separate views and subviews. If provided,
            and split_views=True, each submodule will be placed inside the
            folder of the parent view.

    Returns
    -------
    pysd.model

    """
    model_file_suffix = Path(model_file).suffix.lower()
    if model_file_suffix in vensim_extensions:
        print("\nTranslating model file...\n")
        return pysd.read_vensim(model_file, initialize=False,
                                data_files=data_files,
                                missing_values=missing_values,
                                split_views=split_views, **kwargs)
    elif model_file_suffix in xmile_extensions:
        print("\nTranslating model file...\n")
        return pysd.read_xmile(model_file, initialize=False,
                               data_files=data_files,
                               missing_values=missing_values)
    else:
        return pysd.load(model_file, initialize=False,
                         data_files=data_files,
                         missing_values=missing_values)


def create_configuration(model, options):
    """
    create configuration dict to pass to the run method.

    Parameters
    ----------
    model: pysd.model object

    options: argparse.Namespace

    Returns
    -------
    conf_dict: dict

    """
    conf_dict = {
        "progress": options.progress,
        "params": options.new_values["param"],
        "initial_condition": (options.initial_time or model.time(),
                              options.new_values["initial"]),
        "return_columns": options.return_columns,
        "final_time": options.final_time,
        "time_step": options.time_step,
        "saveper": options.saveper,
        "flatten_output": True,  # need to return totally flat DF
        "return_timestamps": options.return_timestamps,  # given or None,
        "output_file": options.output_file
    }

    if options.import_file:
        conf_dict["initial_condition"] = options.import_file

    return conf_dict



import re
import warnings
import uuid

import parsimonious
from typing import Dict
from pathlib import Path
from chardet import detect


supported_extensions = [".mdl"]


class Grammar():
    _common_grammar = None
    _grammar_path: Path = Path(__file__).parent.joinpath("parsing_grammars")
    _grammar: Dict = {}

    @classmethod
    def get(cls, grammar: str, subs: dict = {}) -> parsimonious.Grammar:
        """Get parsimonious grammar for parsing"""
        if grammar not in cls._grammar:
            # include grammar in the class singleton
            cls._grammar[grammar] = parsimonious.Grammar(
                cls._read_grammar(grammar) % subs
            )

        return cls._grammar[grammar]

    @classmethod
    def _read_grammar(cls, grammar: str) -> str:
        """Read grammar from a file and include common grammar"""
        with cls._gpath(grammar).open(encoding="ascii") as gfile:
            source_grammar: str = gfile.read()

        return cls._include_common_grammar(source_grammar)

    @classmethod
    def _include_common_grammar(cls, source_grammar: str) -> str:
        """Include common grammar"""
        if not cls._common_grammar:
            with cls._gpath("common_grammar").open(encoding="ascii") as gfile:
                cls._common_grammar: str = gfile.read()

        return r"{source_grammar}{common_grammar}".format(
            source_grammar=source_grammar, common_grammar=cls._common_grammar
        )

    @classmethod
    def _gpath(cls, grammar: str) -> Path:
        """Get the grammar file path"""
        return cls._grammar_path.joinpath(grammar).with_suffix(".peg")


def split_arithmetic(structure: object, parsing_ops: dict,
                     expression: str, elements: dict,
                     negatives: set = set()) -> object:
    """
    Split arithmetic pattern and return the corresponding object.

    Parameters
    ----------
    structure: callable
       Callable that generates the arithmetic object to return.
    parsing_ops: dict
       The parsing operators dictionary.
    expression: str
       Original expression with the operator and the hex code to the objects.
    elements: dict
       Dictionary of the hex identifiers and the objects that represent.
    negative: set
       Set of element hex values that must change their sign.

    Returns
    -------
    object: structure
        Final object of the arithmetic operation or initial object if
        no operations are performed.

    """
    pattern = re.compile(parsing_ops)
    parts = pattern.split(expression)  # list of elements ids
    ops = pattern.findall(expression)  # operators list
    if not ops:
        # no operators return original object
        if parts[0] in negatives:
            # make original object negative
            negatives.remove(parts[0])
            return add_element(
                elements,
                structure(["negative"], (elements[parts[0]],)))
        else:
            return expression
    else:
        if not negatives:
            # create arithmetic object
            return add_element(
                elements,
                structure(
                    ops,
                    tuple([elements[id] for id in parts])))
        else:
            # manage negative expressions
            current_id = parts.pop()
            current = elements[current_id]
            if current_id in negatives:
                negatives.remove(current_id)
                current = structure(["negative"], (current,))
            while ops:
                current_id = parts.pop()
                current = structure(
                    [ops.pop()],
                    (elements[current_id], current))
                if current_id in negatives:
                    negatives.remove(current_id)
                    current = structure(["negative"], (current,))

            return add_element(elements, current)


def add_element(elements: dict, element: object) -> str:
    """
    Add element to elements dict using an unique hex identifier

    Parameters
    ----------
    elements: dict
      Dictionary of all elements.

    element: object
      Element to add.

    Returns
    -------
    id: str (hex)
      The name of the key where element is saved in elements.

    """
    id = uuid.uuid4().hex
    elements[id] = element
    return id


def _detect_encoding_from_file(mdl_file: Path) -> str:
    """Detect and return the encoding from a Vensim file"""
    try:
        with mdl_file.open("rb") as in_file:
            f_line: bytes = in_file.readline()
        f_line: str = f_line.decode(detect(f_line)['encoding'])
        return re.search(r"(?<={)(.*)(?=})", f_line).group()
    except (AttributeError, UnicodeDecodeError):
        warnings.warn(
            "No encoding specified or detected to translate the model "
            "file. 'UTF-8' encoding will be used.")
        return "UTF-8"
